# This file generates a GLEW include file suitable for use with Swig.
# It scans the original GLEW file and replaces the defines of extension
# functions with "real" function declarations as needed by Swig for wrapping.
# It also creates a __glFuncAvail function for each OpenGL function, which can
# be used to check the availability of an extension function in the used
# OpenGL driver.
#
# For example the script uses the following two lines to extract the 
# return value, the parameter list and the function name:
#
#    typedef void (GLAPIENTRY * PFNGLACTIVETEXTUREPROC) (GLenum texture);
#    #define glActiveTexture GLEW_GET_FUN(__glewActiveTexture)
#
# These lines are tranformed into the following statements: 
#
#    %{
#    static int __glActiveTextureAvail (void) { return (__glewActiveTexture != NULL); }
#    %}
#    void glActiveTexture (GLenum texture);
#    int  __glActiveTextureAvail (void);
#
# See also the ReadmeGlew.txt file in directory GLSpec.

set inFile  "GL/glew.h"
set swigFile "swigfiles/glewautogen.i"
set helpFile "tclfiles/tcl3dOglHelp.tcl"

puts "Processing GLEW file $inFile ..."
set inFp  [open $inFile "r"]

set foundTypedef false

while { [gets $inFp line] >= 0 } {
    if { [string match "/\**-----*" $line] } {
        set marker [string trim $line "/* -"]
        if { [string match "GL_*" $marker] } {
            set curVersOrExt $marker
        }
    } elseif { [string match "typedef*PFNGL*" $line] } {
        # Found a typedef line.
        set foundTypedef true

        # Extract the return value.
        scan $line "typedef %s" retVal

        # Extract the function name with all letters uppercase.
        set startInd [string first "PFNGL" $line]
        set pfStr [string range $line $startInd end]
        scan $pfStr "PFN%s" funcName
        set startInd [string first "PROC" $funcName]
        set funcName [string range $funcName 0 [expr $startInd -1]]

        # Extract the parameter list.
        set startInd [string first "(" $pfStr]
        set params [string range $pfStr $startInd end]

        # Store the parameter list and the return value in arrays.
        # Key is the uppercase function name.
        set paramList($funcName) $params
        set retList($funcName)   $retVal

    } elseif { [string match "#define*" $line] } {
        # Found a define line.
        # We need to check, if it is a define of a function specified earlier
        # with a typedef line.
        set foundDefine false
        if { $foundTypedef } {
            # Extract the real function name.
            scan $line "#define %s" realFuncName
            set funcName [string toupper $realFuncName]
            if { [info exists paramList($funcName)] } {
                # The name specified in the define line is a function name.
                # Save the SWIG-ready function declaration.
                lappend glExtFuncList $realFuncName
                lappend glExtDefList  "$retList($funcName) $realFuncName $paramList($funcName)"
                lappend glExtVersList "$curVersOrExt"

                # Extract the GLEW function pointer name.
                set startInd [string first "(" $line]
                set glewFuncPtr [string range $line $startInd end]
                set glewFuncPtr [string trim $glewFuncPtr " )("]
                lappend glExtFuncPtr $glewFuncPtr

                set foundDefine true
            }
        }
        if { ! $foundDefine } {
            # If the define was not a function, search for the GLenum definitions.
            if { [string match "#define GL_*" $line] } {
                lappend glEnumList "$line"
            }
        }
    } elseif { [string match "GLAPI *" $line] } {
        # These are the GL 1.1 version available without GetProcAdress.
        set startInd [string first "GLAPIENTRY" $line]
        set startInd [expr [string wordend $line $startInd] + 1]
        set endInd   [string wordend $line $startInd]

        set defStr [string trim [string map {"GLAPIENTRY" "" "GLAPI" ""} $line]]

        lappend glCoreFuncList [string trim [string range $line $startInd $endInd]]
        lappend glCoreDefList  "$defStr"
        lappend glCoreVersList "$curVersOrExt"
    }
}
close $inFp


puts "  Writing Tcl helper file $helpFile ..."
# Open the tcl3dOglHelp.tcl file.
set helpFp [open $helpFile "w"]
puts $helpFp "# This is an automatically generated file. Do not edit.\n"

lappend glAllFuncList $glCoreFuncList $glExtFuncList
puts $helpFp "# Return a list of the names of all wrapped OpenGL functions."
puts $helpFp "proc tcl3dOglGetFuncList {} {\nreturn \[list \\"
foreach glFunc [join $glAllFuncList] {
    puts $helpFp "  $glFunc \\"
}
puts $helpFp "\]\n}\n"

lappend glAllDefList $glCoreDefList $glExtDefList
puts $helpFp "# Return a list of the C-signatures of all wrapped OpenGL functions."
puts $helpFp "proc tcl3dOglGetFuncSignatureList {} {\nreturn \[list \\"
foreach glDef [join $glAllDefList] {
    set sig [string trim $glDef ";"]
    set sig [string map {"[" "\\[" "]" "\\]"} $sig]
    puts $helpFp "  \"$sig\" \\"
}
puts $helpFp "\]\n}\n"

lappend glAllVersList $glCoreVersList $glExtVersList
puts $helpFp "# Return a list of the OpenGL versions or extensions of all wrapped OpenGL functions."
puts $helpFp "proc tcl3dOglGetFuncVersionList {} {\nreturn \[list \\"
foreach glVers [join $glAllVersList] {
    puts $helpFp "  $glVers \\"
}
puts $helpFp "\]\n}\n"
close $helpFp

puts "  Writing SWIG wrapper file $swigFile ..."
# Open the SWIG glewautogen.i file.
set swigFp [open $swigFile "w"]
puts $swigFp "/* This is an automatically generated file. Do not edit. */\n"

# Write the definitions of the functions checking for the existence of an 
# OpenGL extension function.
puts $swigFp "%\{"

puts $swigFp "#include \"GL/glew.h\""

# All core functions (OpenGL 1.1) are available in any case.
foreach coreFunc $glCoreFuncList {
    set defStr [format "static int __%sAvail (void) \{ return 1; \}" $coreFunc]
    puts $swigFp $defStr
}

# For functions introduced after OpenGL 1.1 check the GLEW generated function pointer.
foreach extFunc $glExtFuncList funcPtr $glExtFuncPtr {
    set bodyStr [format "return (%s != NULL)" $funcPtr]
    set defStr [format "static int __%sAvail (void) \{ %s; \}" $extFunc $bodyStr]
    puts $swigFp $defStr
}

puts $swigFp "%\}"

# Write the declarations of all constants and functions being wrapped by SWIG.
puts $swigFp "/* OpenGL enum definitions */"
foreach enumDef $glEnumList {
    puts $swigFp $enumDef
}

puts $swigFp "/* Wrapper for OpenGL core functions in version 1.1 */"
foreach coreFunc $glCoreFuncList coreDef $glCoreDefList coreVers $glCoreVersList {
    puts $swigFp $coreDef
}

puts $swigFp ""
puts $swigFp "/* Wrapper for OpenGL extension functions above version 1.1 */"
foreach extFunc $glExtFuncList extDef $glExtDefList extVers $glExtVersList {
    puts $swigFp $extDef
}

foreach coreFunc $glCoreFuncList {
    set defStr [format "static int __%sAvail (void);" $coreFunc]
    puts $swigFp $defStr
}

foreach extFunc $glExtFuncList {
    set defStr [format "int __%sAvail (void);" $extFunc]
    puts $swigFp $defStr
}

# Add the GLEW initialization function.
puts $swigFp "GLenum glewInit ();"
puts $swigFp "const char* glewGetString (GLenum name);"

close $swigFp

puts "Done."
exit 0
