/******************************************************************************
 *{@C
 *      Copyright:      2005-2008 Paul Obermeier (obermeier@tcl3d.org)
 *
 *                      See the file "Tcl3D_License.txt" for information on
 *                      usage and redistribution of this file, and for a
 *                      DISCLAIMER OF ALL WARRANTIES.
 *
 *      Module:         Tcl3D -> tcl3dOgl
 *      Filename:       tcl3dShapesMisc.c
 *
 *      Author:         Paul Obermeier
 *
 *      Description:    C functions for generating miscellaneous standard 
 *                      shapes. See also the GLUT shape functions.
 *                      Currently the following shapes are implemented:
 *                      tcl3dCube
 *                      tcl3dSphere
 *                      tcl3dHelix
 *
 *****************************************************************************/

#if defined(WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
#include <winnt.h>
#endif

#include <math.h>
#include <GL/glew.h>

#include "tcl3dVecMath.h"
#include "tcl3dShapesMisc.h"

/*
 * Draw a textured box with style "type". Type can be GL_QUADS or GL_LINE_LOOPS.
 * The box is specified by the lower-left and upper-right corners.
 */

void tcl3dBox (const float lowerLeft[3], const float upperRight[3], GLenum type)
{
    float llx = lowerLeft[0];
    float lly = lowerLeft[1];
    float llz = lowerLeft[2];
    float urx = upperRight[0];
    float ury = upperRight[1];
    float urz = upperRight[2];

    /* Front Face */
    glBegin (type);
    glNormal3f (0.0f, 0.0f, 0.5f);
    glTexCoord2f (0.0f, 0.0f); glVertex3f (llx, lly, urz);
    glTexCoord2f (1.0f, 0.0f); glVertex3f (urx, lly, urz);
    glTexCoord2f (1.0f, 1.0f); glVertex3f (urx, ury, urz);
    glTexCoord2f (0.0f, 1.0f); glVertex3f (llx, ury, urz);
    glEnd ();
    /* Back Face */
    glBegin (type);
    glNormal3f (0.0f, 0.0f, -0.5f);
    glTexCoord2f (1.0f, 0.0f); glVertex3f (llx, lly, llz);
    glTexCoord2f (1.0f, 1.0f); glVertex3f (llx, ury, llz);
    glTexCoord2f (0.0f, 1.0f); glVertex3f (urx, ury, llz);
    glTexCoord2f (0.0f, 0.0f); glVertex3f (urx, lly, llz);
    glEnd ();
    /* Top Face */
    glBegin (type);
    glNormal3f (0.0f, 0.5f, 0.0f);
    glTexCoord2f (0.0f, 1.0f); glVertex3f (llx, ury, llz);
    glTexCoord2f (0.0f, 0.0f); glVertex3f (llx, ury, urz);
    glTexCoord2f (1.0f, 0.0f); glVertex3f (urx, ury, urz);
    glTexCoord2f (1.0f, 1.0f); glVertex3f (urx, ury, llz);
    glEnd ();
    /* Bottom Face */
    glBegin (type);
    glNormal3f (0.0f, -0.5f, 0.0f);
    glTexCoord2f (1.0f, 1.0f); glVertex3f (llx, lly, llz);
    glTexCoord2f (0.0f, 1.0f); glVertex3f (urx, lly, llz);
    glTexCoord2f (0.0f, 0.0f); glVertex3f (urx, lly, urz);
    glTexCoord2f (1.0f, 0.0f); glVertex3f (llx, lly, urz);
    glEnd ();
    /* Right Face */
    glBegin (type);
    glNormal3f (0.5f, 0.0f, 0.0f);
    glTexCoord2f (1.0f, 0.0f); glVertex3f (urx, lly, llz);
    glTexCoord2f (1.0f, 1.0f); glVertex3f (urx, ury, llz);
    glTexCoord2f (0.0f, 1.0f); glVertex3f (urx, ury, urz);
    glTexCoord2f (0.0f, 0.0f); glVertex3f (urx, lly, urz);
    glEnd ();
    /* Left Face */
    glBegin (type);
    glNormal3f (-0.5f, 0.0f, 0.0f);
    glTexCoord2f (0.0f, 0.0f); glVertex3f (llx, lly, llz);
    glTexCoord2f (1.0f, 0.0f); glVertex3f (llx, lly, urz);
    glTexCoord2f (1.0f, 1.0f); glVertex3f (llx, ury, urz);
    glTexCoord2f (0.0f, 1.0f); glVertex3f (llx, ury, llz);
    glEnd ();
}

/*
 * Draw a textured cube as 6 GL_QUADS.
 * The center of the cube is at (cx, cy, cz) and it's size is "size".
 */
void tcl3dCube (float cx, float cy, float cz, float size)
{
    float lowerLeft[3];
    float upperRight[3];

    lowerLeft[0]  = cx - 0.5 * size;
    lowerLeft[1]  = cy - 0.5 * size;
    lowerLeft[2]  = cz - 0.5 * size;
    upperRight[0] = cx + 0.5 * size;
    upperRight[1] = cy + 0.5 * size;
    upperRight[2] = cz + 0.5 * size;

    tcl3dBox (lowerLeft, upperRight, GL_QUADS);
}

/*
 * Create a sphere centered at cy, cx, cz with radius r, and 
 * precision p.
 * Based on a function written by Paul Bourke.
 * http://astronomy.swin.edu.au/~pbourke/opengl/sphere/
 */
void tcl3dSphere (float cx, float cy, float cz, float r, int p)
{
    int i, j;

    const float PI     = 3.14159265358979f;
    const float TWOPI  = 6.28318530717958f;
    const float PIDIV2 = 1.57079632679489f;

    float theta1 = 0.0;
    float theta2 = 0.0;
    float theta3 = 0.0;

    float ex = 0.0f;
    float ey = 0.0f;
    float ez = 0.0f;

    float px = 0.0f;
    float py = 0.0f;
    float pz = 0.0f;

    /* Disallow a negative number for radius. */
    if ( r < 0 ) {
        r = -r;
    }

    /* Disallow a negative number for precision. */
    if ( p < 0 ) {
        p = -p;
    }

    /* If the sphere is too small, just render a OpenGL point instead. */
    if ( p < 4 || r <= 0 ) {
        glBegin (GL_POINTS);
        glVertex3f (cx, cy, cz);
        glEnd ();
        return;
    }

    for (i=0; i<p/2; ++i) {
        theta1 = i * TWOPI / p - PIDIV2;
        theta2 = (i + 1) * TWOPI / p - PIDIV2;

        glBegin (GL_TRIANGLE_STRIP);
        for (j=0; j<=p; ++j) {
            theta3 = j * TWOPI / p;

            ex = cos(theta2) * cos(theta3);
            ey = sin(theta2);
            ez = cos(theta2) * sin(theta3);
            px = cx + r * ex;
            py = cy + r * ey;
            pz = cz + r * ez;

            glNormal3f (ex, ey, ez);
            glTexCoord2f (-(j/(float)p) , 2*(i+1)/(float)p);
            glVertex3f (px, py, pz);

            ex = cos(theta1) * cos(theta3);
            ey = sin(theta1);
            ez = cos(theta1) * sin(theta3);
            px = cx + r * ex;
            py = cy + r * ey;
            pz = cz + r * ez;

            glNormal3f (ex, ey, ez);
            glTexCoord2f (-(j/(float)p), 2*i/(float)p);
            glVertex3f (px, py, pz);
        }
        glEnd ();
    }
}

/*
 * Draw a helix.
 * This code is based on an excerpt of NeHe's tutorial lesson 36
 * written by Dario Corno.
 */
void tcl3dHelix ( float cx, float cy, float cz, float r, int twists ) {
    const float PI = 3.14159265358979f;
    float x, y, z;
    float phi, theta;
    float u, v;
    float vtx0[3], vtx1[3], vtx2[3], vtx3[3];
    float normal[3];

    glBegin (GL_QUADS);
    for (phi=0.0f; phi<=360.0f; phi+=20.0f) {
        for (theta=0.0f; theta<=360.0f*twists; theta+=20.0f) {
            /* Calculate angle of first point (0, 0) */
            v=(phi/180.0f*PI);
            u=(theta/180.0f*PI);

            /* Calculate position of first point */
            x=(cos(u)*(2.0f+cos(v)))*r;
            y=(sin(u)*(2.0f+cos(v)))*r;
            z=((( u-(2.0f*PI)) + sin(v) ) * r);

            /* Set value of first vertex */
            vtx0[0]=x;
            vtx0[1]=y;
            vtx0[2]=z;

            /* Calculate angle of second point (0, 20) */
            v=(phi/180.0f*PI);
            u=((theta+20)/180.0f*PI);

            /* Calculate position of 2nd point */
            x=(cos(u)*(2.0f+cos(v) ))*r;
            y=(sin(u)*(2.0f+cos(v) ))*r;
            z=((( u-(2.0f*PI)) + sin(v) ) * r);

            /* Set value of second vertex */
            vtx1[0]=x;
            vtx1[1]=y;
            vtx1[2]=z;

            /* Calculate angle of third point (20, 20) */
            v=((phi+20)/180.0f*PI);
            u=((theta+20)/180.0f*PI);

            /* Calculate position of third point */
            x=(cos(u)*(2.0f+cos(v) ))*r;
            y=(sin(u)*(2.0f+cos(v) ))*r;
            z=((( u-(2.0f*PI)) + sin(v) ) * r);

            /* Set value of third vertex */
            vtx2[0]=x;
            vtx2[1]=y;
            vtx2[2]=z;

            /* Calculate angle of fourth point (20, 0) */
            v=((phi+20)/180.0f*PI);
            u=((theta)/180.0f*PI);

            /* Calculate position of fourth point */
            x=(cos(u)*(2.0f+cos(v) ))*r;
            y=(sin(u)*(2.0f+cos(v) ))*r;
            z=((( u-(2.0f*PI)) + sin(v) ) * r);

            /* Set value of fourth vertex */
            vtx3[0]=x;
            vtx3[1]=y;
            vtx3[2]=z;

            /* Calculate the quad normal and set it. */
            tcl3dVec3fPlaneNormal (vtx0, vtx1, vtx2, normal);
            glNormal3f (normal[0], normal[1], normal[2]);

            /* Render The Quad */
            glVertex3f (vtx0[0] + cx, vtx0[1] + cy, vtx0[2] + cz);
            glVertex3f (vtx1[0] + cx, vtx1[1] + cy, vtx1[2] + cz);
            glVertex3f (vtx2[0] + cx, vtx2[1] + cy, vtx2[2] + cz);
            glVertex3f (vtx3[0] + cx, vtx3[1] + cy, vtx3[2] + cz);
        }
    }
    glEnd ();
}
