/******************************************************************************
 *{@C
 *      Copyright:      2005-2008 Paul Obermeier (obermeier@tcl3d.org)
 *
 *                      See the file "Tcl3D_License.txt" for information on
 *                      usage and redistribution of this file, and for a
 *                      DISCLAIMER OF ALL WARRANTIES.
 *
 *      Module:         Tcl3D -> tcl3dOgl
 *      Filename:       tcl3dVecMath.c
 *
 *      Author:         Paul Obermeier
 *
 *      Description:    C functions for handling 3D vectors and transformation
 *                      matrices in OpenGL compatible style.
 *                      This module is intended to be wrapped with Swig for the
 *                      Tcl3D package. Vectors and matrices are represented
 *                      as standard Tcl lists.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "tcl3dVecMath.h"

/*
    MIN_NRM_FLOAT : The smallest normalized Float32 value greater than zero

    The assertion for normalized numbers is:
    1.0f + MIN_NRM_FLOAT != 1.0f
*/

#ifndef FLT_MIN
    #define FLT_MIN 1.175494351e-38F
#endif
#ifndef FLT_EPSILON
    #define FLT_EPSILON 1.192092896e-07F
#endif

#define MIN_NRM_FLOAT (FLT_MIN * (1.0f / FLT_EPSILON))

#define DEGTORAD(degree) ((degree) * (3.141592654f / 180.0f))
#define RADTODEG(radian) ((radian) * (180.0f / 3.141592654f))
#define ABS(a) ((a)>0?(a):(-(a)))

#define X 0
#define Y 1
#define Z 2

/**
  * Check, if two vectors are equal.
  * Each vector component of v1 must not differ by more than
  * delta units from the corresponding component of v2.
  */
int tcl3dVec3fCompare (float *v1, float *v2, float delta)
{
    return ABS(v2[0] - v1[0]) < delta &&
           ABS(v2[1] - v1[1]) < delta &&
           ABS(v2[2] - v1[2]) < delta;
}

/**
  * Initialize a 3D vector with zeros.
  */
void tcl3dVec3fIdentity (float *res)
{
    res[0] = 0.0f;
    res[1] = 0.0f;
    res[2] = 0.0f;
    return;
}

/**
  * Copy a 3D vector: res = v.
  */
void tcl3dVec3fCopy (float *v, float *res)
{
    res[0] = v[0];
    res[1] = v[1];
    res[2] = v[2];
    return;
}

/**
  * Return the length of a 3D vector.
  */
float tcl3dVec3fLength (float *v)
{
    return ((float)sqrt (v[X]*v[X] + v[Y]*v[Y] + v[Z]*v[Z]));
}

/**
  * Scale a 3D vector so that its length becomes 1.0:
  *     v = 1 / max (MIN_NRM_FLOAT, length(v)) * v
  * True (1), if length(v) was greater than MIN_NRM_FLOAT
  * before v was scaled, otherwise false (0).
  */
int tcl3dVec3fNormalize (float *v)
{
    float t;

    /* To avoid numerical difficulties, do not replace the divisions
       by multiplications! */

    if ((t = sqrt (v[0] * v[0] + v[1] * v[1] + v[2] * v[2])) < MIN_NRM_FLOAT)
    {
        v[0] /= MIN_NRM_FLOAT;
        v[1] /= MIN_NRM_FLOAT;
        v[2] /= MIN_NRM_FLOAT;
        return 0;
    }
    v[0] /= t;
    v[1] /= t;
    v[2] /= t;
    return 1;
}

/**
  * Return the distance between two 3D points.
  */
float tcl3dVec3fDistance (float *v1,  float *v2)
{
    float t0, t1, t2;

    t0 = v1[0] - v2[0];
    t1 = v1[1] - v2[1];
    t2 = v1[2] - v2[2];
    return sqrt (t0 * t0 + t1 * t1 + t2 * t2);
}

/**
  * Return the dot product of two 3D vectors.
  */
float tcl3dVec3fDotProduct (float *v1,  float *v2)
{
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}

/**
  * Calculate the cross product of two 3D vectors: res = v1 x v2.
  * "res" must not refer to the same memory location as "v1" or "v2".
  */
void tcl3dVec3fCrossProduct (float *v1, float *v2, float *res)
{
    res[0] = v1[1] * v2[2] - v2[1] * v1[2];
    res[1] = v1[2] * v2[0] - v2[2] * v1[0];
    res[2] = v1[0] * v2[1] - v2[0] * v1[1];
}

/**
  * Add two 3D vectors: res = v1 + v2.
  */
void tcl3dVec3fAdd (float *v1, float *v2, float *res)
{
    res[0] = v1[0] + v2[0];
    res[1] = v1[1] + v2[1];
    res[2] = v1[2] + v2[2];
    return;
}

/**
  * Subtract two 3D vectors: res = v1 - v2.
  */
void tcl3dVec3fSubtract (float *v1, float *v2, float *res)
{
    res[0] = v1[0] - v2[0];
    res[1] = v1[1] - v2[1];
    res[2] = v1[2] - v2[2];
    return;
}

/**
  * Multiply a 3D vector with a scalar: res = v * scalar.
  */
void tcl3dVec3fScale (float scalar, float *v, float *res)
{
    res[0] = v[0] * scalar;
    res[1] = v[1] * scalar;
    res[2] = v[2] * scalar;
    return;
}

/**
  * Create the normal of the plane defined by three points.
  */
void tcl3dVec3fPlaneNormal (float *p1, float *p2, float *p3, float *res)
{
    float v1[3], v2[3];

    /* Find the vector between 2 points by subtracting
       the x,y,z coordinates from one point to another. */

    /* Calculate the vector from point 2 to point 1 */
    tcl3dVec3fSubtract (p1, p2, v1);

    /* Calculate the vector from point 3 to point 2 */
    tcl3dVec3fSubtract (p2, p3, v2);

    /* Compute the cross product to give us the surface normal */
    tcl3dVec3fCrossProduct (v1, v2, res);

    /* Normalize the surface normal vector */
    tcl3dVec3fNormalize (res);
}

/**
  * Check, if two matrices are equal.
  * Each matrix component of m1 must not differ by more than
  * delta units from the corresponding component of m2.
  */
int tcl3dMatfCompare (float *m1, float *m2, float delta)
{
    return ABS(m2[0]  - m1[0]) < delta &&
           ABS(m2[1]  - m1[1]) < delta &&
           ABS(m2[2]  - m1[2]) < delta &&
           ABS(m2[3]  - m1[3]) < delta &&
           ABS(m2[4]  - m1[4]) < delta &&
           ABS(m2[5]  - m1[5]) < delta &&
           ABS(m2[6]  - m1[6]) < delta &&
           ABS(m2[7]  - m1[7]) < delta &&
           ABS(m2[8]  - m1[8]) < delta &&
           ABS(m2[9]  - m1[9]) < delta &&
           ABS(m2[10] - m1[10]) < delta &&
           ABS(m2[11] - m1[11]) < delta &&
           ABS(m2[12] - m1[12]) < delta &&
           ABS(m2[13] - m1[13]) < delta &&
           ABS(m2[14] - m1[14]) < delta &&
           ABS(m2[15] - m1[15]) < delta;
}

/**
  * Build the identity transformation matrix.
  */
void tcl3dMatfIdentity (float *res)
{
    res[0]=1.0f; res[4]=0.0f; res[8] =0.0f; res[12]=0.0f;
    res[1]=0.0f; res[5]=1.0f; res[9] =0.0f; res[13]=0.0f;
    res[2]=0.0f; res[6]=0.0f; res[10]=1.0f; res[14]=0.0f;
    res[3]=0.0f; res[7]=0.0f; res[11]=0.0f; res[15]=1.0f;
    return;
}

/**
  * Copy a transformation matrix: res = m.
  */
void tcl3dMatfCopy (float *m, float *res)
{
    res[0]  = m[0];  res[1]  = m[1];  res[2]  = m[2];  res[3]  = m[3];
    res[4]  = m[4];  res[5]  = m[5];  res[6]  = m[6];  res[7]  = m[7];
    res[8]  = m[8];  res[9]  = m[9];  res[10] = m[10]; res[11] = m[11];
    res[12] = m[12]; res[13] = m[13]; res[14] = m[14]; res[15] = m[15];
    return;
}

/**
  *  Build a translation matrix based on a 3D vector.
  */
void tcl3dMatfTranslatev (float *v, float *m)
{
    tcl3dMatfIdentity (m);

    m[12] = v[0];
    m[13] = v[1];
    m[14] = v[2];
}

/**
  * Build a translation matrix based on 3 scalar values.
  */
void tcl3dMatfTranslate (float tx, float ty, float tz, float *m)
{
    tcl3dMatfIdentity (m);

    m[12] = tx;
    m[13] = ty;
    m[14] = tz;
}

/**
  * Build a rotation matrix based on angle and axis.
  * Angle is given in degrees.
  */
void tcl3dMatfRotate (float angle, float *axis, float *m)
{
    float s, c, v;
    float ux, uy, uz;

    s = sin(DEGTORAD(angle));
    c = cos(DEGTORAD(angle));
    v = 1-c;

    tcl3dVec3fNormalize (axis);

    ux = axis[0];
    uy = axis[1];
    uz = axis[2];
    
    m[0]  = ux*ux*v + c;
    m[1]  = ux*uy*v + uz*s;
    m[2]  = uz*ux*v - uy*s;
    m[3]  = 0.0f;
    
    m[4]  = ux*uy*v - uz*s;
    m[5]  = uy*uy*v + c;
    m[6]  = uz*uy*v + ux*s;
    m[7]  = 0.0f;
    
    m[8]  = ux*uz*v + uy*s;
    m[9]  = uz*uy*v - ux*s;
    m[10] = uz*uz*v + c;
    m[11] = 0.0f;
    
    m[12] = 0.0f;
    m[13] = 0.0f;
    m[14] = 0.0f;
    m[15] = 1.0f;
}

/**
  * Build a rotation matrix based on angle around x axis.
  * Angle is given in degrees.
  */
void tcl3dMatfRotateX (float angle, float *m)
{
    float s, c;

    s = sin(DEGTORAD(angle));
    c = cos(DEGTORAD(angle));

    tcl3dMatfIdentity (m);

    m[5]  =  c;
    m[6]  =  s;
    m[9]  = -s;
    m[10] =  c;
}

/**
  * Build a rotation matrix based on angle around y axis.
  * Angle is given in degrees.
  */
void tcl3dMatfRotateY (float angle, float *m)
{
    float s, c;

    s = sin(DEGTORAD(angle));
    c = cos(DEGTORAD(angle));

    tcl3dMatfIdentity (m);

    m[0]  =  c;
    m[2]  = -s;
    m[8]  =  s;
    m[10] =  c;
}

/**
  * Build a rotation matrix based on angle around z axis.
  * Angle is given in degrees.
  */
void tcl3dMatfRotateZ (float angle, float *m)
{
    float s, c;

    s = sin(DEGTORAD(angle));
    c = cos(DEGTORAD(angle));

    tcl3dMatfIdentity (m);

    m[0] =  c;
    m[1] =  s;
    m[4] = -s;
    m[5] =  c;
}

/**
  * Build a scale matrix based on a 3D vector.
  */
void tcl3dMatfScalev (float *v, float *m)
{
    tcl3dMatfIdentity (m);

    m[0]  = v[0];
    m[5]  = v[1];
    m[10] = v[2];
}

/**
  * Build a scale matrix based on 3 scalar values.
  */
void tcl3dMatfScale (float sx, float sy, float sz, float *m)
{
    tcl3dMatfIdentity (m);

    m[0]  = sx;
    m[5]  = sy;
    m[10] = sz;
}

/**
  * Transform a point by a given matrix.
  */
void tcl3dMatfTransformPoint (float *v, float *m, float *res)
{
    float x, y, z;

    x = v[0];
    y = v[1];
    z = v[2];

    res[0] = x * m[0] +
             y * m[4] +
             z * m[8] + m[12];

    res[1] = x * m[1] +
             y * m[5] +
             z * m[9] + m[13];

    res[2] = x * m[2] +
             y * m[6] +
             z * m[10]+ m[14];
}

/**
  * Transform a 3D vector by a given matrix.
  */
void tcl3dMatfTransformVector (float *v, float *m, float *res)
{
    float x, y, z;

    x = v[0];
    y = v[1];
    z = v[2];

    res[0] = x * m[0] +
             y * m[4] +
             z * m[8];

    res[1] = x * m[1] +
             y * m[5] +
             z * m[9];

    res[2] = x * m[2] +
             y * m[6] +
             z * m[10];
}

/**
  * Multiply two transformation matrices.
  */
void tcl3dMatfMult (float *m1, float *m2, float *res)
{
    res[0]  = (m1[0]*m2[0])+(m1[4]*m2[1])+(m1[8]*m2[2])+(m1[12]*m2[3]);
    res[1]  = (m1[1]*m2[0])+(m1[5]*m2[1])+(m1[9]*m2[2])+(m1[13]*m2[3]);
    res[2]  = (m1[2]*m2[0])+(m1[6]*m2[1])+(m1[10]*m2[2])+(m1[14]*m2[3]);
    res[3]  = (m1[3]*m2[0])+(m1[7]*m2[1])+(m1[11]*m2[2])+(m1[15]*m2[3]);

    res[4]  = (m1[0]*m2[4])+(m1[4]*m2[5])+(m1[8]*m2[6])+(m1[12]*m2[7]);
    res[5]  = (m1[1]*m2[4])+(m1[5]*m2[5])+(m1[9]*m2[6])+(m1[13]*m2[7]);
    res[6]  = (m1[2]*m2[4])+(m1[6]*m2[5])+(m1[10]*m2[6])+(m1[14]*m2[7]);
    res[7]  = (m1[3]*m2[4])+(m1[7]*m2[5])+(m1[11]*m2[6])+(m1[15]*m2[7]);

    res[8]  = (m1[0]*m2[8])+(m1[4]*m2[9])+(m1[8]*m2[10])+(m1[12]*m2[11]);
    res[9]  = (m1[1]*m2[8])+(m1[5]*m2[9])+(m1[9]*m2[10])+(m1[13]*m2[11]);
    res[10] = (m1[2]*m2[8])+(m1[6]*m2[9])+(m1[10]*m2[10])+(m1[14]*m2[11]);
    res[11] = (m1[3]*m2[8])+(m1[7]*m2[9])+(m1[11]*m2[10])+(m1[15]*m2[11]);

    res[12] = (m1[0]*m2[12])+(m1[4]*m2[13])+(m1[8]*m2[14])+(m1[12]*m2[15]);
    res[13] = (m1[1]*m2[12])+(m1[5]*m2[13])+(m1[9]*m2[14])+(m1[13]*m2[15]);
    res[14] = (m1[2]*m2[12])+(m1[6]*m2[13])+(m1[10]*m2[14])+(m1[14]*m2[15]);
    res[15] = (m1[3]*m2[12])+(m1[7]*m2[13])+(m1[11]*m2[14])+(m1[15]*m2[15]);
}

/**
  * Invert a transformation matrix.
  */
void tcl3dMatfInvert (float *m, float *res)
{
    /*
     OpenGL matrices are column major and can be quite confusing to access 
     when stored in the typical, one-dimensional array often used by the API.
     Here are some shorthand conversion macros, which convert a row/column 
     combination into an array index.
    */
    
    #define MAT(r,c) m[c*4+r]

    #define m11 MAT(0,0)
    #define m12 MAT(0,1)
    #define m13 MAT(0,2)
    #define m14 MAT(0,3)
    #define m21 MAT(1,0)
    #define m22 MAT(1,1)
    #define m23 MAT(1,2)
    #define m24 MAT(1,3)
    #define m31 MAT(2,0)
    #define m32 MAT(2,1)
    #define m33 MAT(2,2)
    #define m34 MAT(2,3)
    #define m41 MAT(3,0)
    #define m42 MAT(3,1)
    #define m43 MAT(3,2)
    #define m44 MAT(3,3)

    /* Inverse = adjoint / det. (See linear algebra texts.)
       pre-compute 2x2 dets for last two rows when computing
       cofactors of first two rows.
     */
    float d12, d13, d23, d24, d34, d41;
    float det, invDet;
    float tmp[16];
    float identity[16] = {
       1.0, 0.0, 0.0, 0.0,
       0.0, 1.0, 0.0, 0.0,
       0.0, 0.0, 1.0, 0.0,
       0.0, 0.0, 0.0, 1.0
    };

    d12 = (m31 * m42 - m41 * m32);
    d13 = (m31 * m43 - m41 * m33);
    d23 = (m32 * m43 - m42 * m33);
    d24 = (m32 * m44 - m42 * m34);
    d34 = (m33 * m44 - m43 * m34);
    d41 = (m34 * m41 - m44 * m31);
    
    tmp[0] =  (m22 * d34 - m23 * d24 + m24 * d23);
    tmp[1] = -(m21 * d34 + m23 * d41 + m24 * d13);
    tmp[2] =  (m21 * d24 + m22 * d41 + m24 * d12);
    tmp[3] = -(m21 * d23 - m22 * d13 + m23 * d12);

    /* Compute determinant as early as possible using these cofactors. */
    det = m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2] + m14 * tmp[3];

    /* Run singularity test. */
    if ( det == 0.0 ) {
        printf ("Warning: Call to invertMatrix produced a Singular matrix.\n");
        memcpy (res, identity, 16*sizeof(float));
    } else {
       invDet = 1.0f / det;
       
       /* Compute rest of inverse. */
       tmp[0] *= invDet;
       tmp[1] *= invDet;
       tmp[2] *= invDet;
       tmp[3] *= invDet;

       tmp[4] = -(m12 * d34 - m13 * d24 + m14 * d23) * invDet;
       tmp[5] =  (m11 * d34 + m13 * d41 + m14 * d13) * invDet;
       tmp[6] = -(m11 * d24 + m12 * d41 + m14 * d12) * invDet;
       tmp[7] =  (m11 * d23 - m12 * d13 + m13 * d12) * invDet;

       /* Pre-compute 2x2 dets for first two rows when computing cofactors 
          of last two rows. */
       d12 = m11 * m22 - m21 * m12;
       d13 = m11 * m23 - m21 * m13;
       d23 = m12 * m23 - m22 * m13;
       d24 = m12 * m24 - m22 * m14;
       d34 = m13 * m24 - m23 * m14;
       d41 = m14 * m21 - m24 * m11;

       tmp[8]  =  (m42 * d34 - m43 * d24 + m44 * d23) * invDet;
       tmp[9]  = -(m41 * d34 + m43 * d41 + m44 * d13) * invDet;
       tmp[10] =  (m41 * d24 + m42 * d41 + m44 * d12) * invDet;
       tmp[11] = -(m41 * d23 - m42 * d13 + m43 * d12) * invDet;
       tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23) * invDet;
       tmp[13] =  (m31 * d34 + m33 * d41 + m34 * d13) * invDet;
       tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12) * invDet;
       tmp[15] =  (m31 * d23 - m32 * d13 + m33 * d12) * invDet;

       memcpy (res, tmp, 16*sizeof(float));
    }

    #undef m11
    #undef m12
    #undef m13
    #undef m14
    #undef m21
    #undef m22
    #undef m23
    #undef m24
    #undef m31
    #undef m32
    #undef m33
    #undef m34
    #undef m41
    #undef m42
    #undef m43
    #undef m44
    #undef MAT
}

/**
  * Transpose a transformation matrix.
  */
void tcl3dMatfTranspose (float *m, float *res)
{
    res[0]  = m[0];
    res[1]  = m[4];
    res[2]  = m[8];
    res[3]  = m[12];

    res[4]  = m[1];
    res[5]  = m[5];
    res[6]  = m[9];
    res[7]  = m[13];

    res[8]  = m[2];
    res[9]  = m[6];
    res[10] = m[10];
    res[11] = m[14];

    res[12] = m[3];
    res[13] = m[7];
    res[14] = m[11];
    res[15] = m[15];
}

/**
  * Decompose a rotation matrix into the 3 Euler angles.
  */
int tcl3dMatfGetAngles1 (float *m, float *res)
{
    res[1] = asin(m[0*4+2]);
    if (fabs(cos(res[1])) > 0.0000001) {
        res[0] = atan2 (m[1*4+2], m[2*4+2]);
        res[2] = atan2 (m[0*4+1], m[0*4+0]);
        return 1;
    } else {
        res[0] = atan2 (m[1*4+0], m[1*4+1]);
        res[0] = 0.0;
        return 0;
    }
}

int tcl3dMatfGetAngles (float *m, float *res)
{
    #define HALF_PI 1.5707963267948966192313216916398
    if ( m[2] < 1.0f ) {
        if ( m[2] > -1.0f ) {
            res[0] = atan2(-m[6],m[10]);
            res[1] = asin(m[2]);
            res[2] = atan2(-m[1],m[0]);
            return 1;
        } else {
            // WARNING.  Not unique.  XA - ZA = -atan2(r10,r11)
            res[0] = -atan2(m[4],m[5]);
            res[1] = -HALF_PI;
            res[2] = 0.0f;
            return 0;
        }
    } else {
        // WARNING.  Not unique.  XAngle + ZAngle = atan2(r10,r11)
        res[0] = atan2(m[4],m[5]);
        res[1] = HALF_PI;
        res[2] = 0.0f;
        return 0;
    }
}

unsigned int tcl3dRoundUpPow2 (unsigned int val)
{
    val--;
    val |= val >>  1;
    val |= val >>  2;
    val |= val >>  4;
    val |= val >>  8;
    val |= val >> 16;
    return val+1;
}

unsigned int tcl3dIsPow2 (unsigned int val) {
    unsigned int i = 1;
    while (i <= val) {
        if (val == i) {
            return 1;
        }
        i *= 2;
    }
    return 0;
}
