#******************************************************************************
#
#       Copyright:      2005-2008 Paul Obermeier (obermeier@tcl3d.org)
#
#                       See the file "Tcl3D_License.txt" for information on
#                       usage and redistribution of this file, and for a
#                       DISCLAIMER OF ALL WARRANTIES.
#
#       Module:         Tcl3D -> tcl3dOgl
#       Filename:       tcl3dOglUtil.tcl
#
#       Author:         Paul Obermeier
#
#       Description:    Tcl module with miscellaneous utility
#                       procedures related to the OpenGL module.
#
#******************************************************************************

###############################################################################
#[@e
#       Name:           tcl3dOglExtInit - Initialize the extension library.
#
#       Synopsis:       tcl3dOglExtInit {}
#
#       Description:    Initialize the OpenGL extension library.
#                       If no OpenGL context has been established (i.e. a Togl
#                       window has not been created), the function throws an 
#                       error.
#                       It is recommended to add a call to tcl3dOglExtInit in
#                       the create callback.
#                       Note: With Tcl3D versions starting at 0.4, this 
#                       procedure is not needed anymore, but you may
#                       leave it in your code for backwards compatibility.
#
#       See also:
#
###############################################################################

# OBSOLETE tcl3dOglExtInit 0.4.0 None
proc tcl3dOglExtInit {} {
    if { [info commands glewInit] ne "" } {
        # GLEW based wrapper. Tcl3D Version >= 0.4
        glewInit
    } elseif { [info commands glexGetVersion] ne "" } {
        # OglExt based wrapper. Tcl3D Version >= 0.2
        glexGetVersion
    } else {
        error "OpenGL extension library not wrapped"
    }
}

# OBSOLETE tcl3dInit 0.3.2 tcl3dOglExtInit
proc tcl3dInit {} {
    tcl3dOglExtInit
}

###############################################################################
#[@e
#       Name:           glMultiDrawElements - Draw multiple array elements.
#
#       Synopsis:       glMultiDrawElements {
#                               mode count type indices primcount }
#
#       Description:    mode      : GLenum
#                       count     : tcl3dVector of type GLuint
#                       type      : GLenum
#                       indices   : List of tcl3dVectors
#                       primcount : integer
#
#                       Tcl wrapper for the OpenGL function glMultiDrawElements.
#                       As the Tcl3D Swig wrapper currently does not support
#                       "void **" pointers, the multiple array elements must be
#                       specified as a list of tcl3dVectors.
#
#       See also:
#
###############################################################################

proc glMultiDrawElements { mode count type indices primcount } {
    for { set i 0 } { $i < $primcount } { incr i } {
        if { [$count get $i] > 0 } {
            glDrawElements $mode [$count get $i] $type [lindex $indices $i]
        }
    }
}

###############################################################################
#[@e
#       Name:           tcl3dOglGetError - Check for an OpenGL error.
#
#       Synopsis:       tcl3dOglGetError {}
#
#       Description:    Check, if an OpenGL related error has been occurred.
#
#                       If no error occurred, an emtpy string is returned.
#                       Otherwise a formatted string showing the error number
#                       and the error message is returned.
#
#       See also:
#
###############################################################################

proc tcl3dOglGetError {} {
    set err [glGetError]

    if { $err == $::GL_NO_ERROR } {
        return ""
    }

    return [format "GL error: 0X%X (%s)" $err [gluErrorString $err]]
}

# OBSOLETE tcl3dCheckGlError 0.3.2 tcl3dOglGetError
proc tcl3dCheckGlError { msg } {
    set err [glGetError]

    if { $err == $::GL_NO_ERROR } {
        return 0
    }

    puts [format "%s: GL error: 0X%X (%s)" $msg $err [gluErrorString $err]]
    return 1
}

###############################################################################
#[@e
#       Name:           tcl3dOglShaderSource - Wrapper for glShaderSource.
#
#       Synopsis:       tcl3dOglShaderSource { shaderId shaderString }
#
#       Description:    shaderId     : Shader handle
#                       shaderString : string
#       
#                       Wrapper for easier use of OpenGL function glShaderSource.
#                       In contrast to glShaderSource only the shader program
#                       identifier (created with a call to glCreateShaderObject)
#                       and the shader source have to be specified.
#
#       See also:
#
###############################################################################

proc tcl3dOglShaderSource { shaderId shaderString } {
    set shaderStringList [list $shaderString]
    set lenList [list [string length $shaderString]]
    glShaderSource $shaderId 1 $shaderStringList $lenList
}
